<!DOCTYPE html>
<html lang="en" class="no-js">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="dns-prefetch" href="//fonts.gstatic.com">
    <link href="https://fonts.googleapis.com/css?family=Inter" rel="stylesheet">
    <title>{{ exc.status_code }}. {{ exc.detail }}</title>
    <style>
        html, body {
            color: #6b7280;
            font-family: 'Inter', sans-serif;
            font-weight: 100;
            height: 100vh;
            margin: 0;
            font-size: 24px;
        }

        h3 {
            font-size: 36px;
            margin-right: 16px;
            font-weight: 600;
        }

        .page {
            height: 100vh;
            width: 100vw;
            display: flex;
            flex: 1;
            position: relative;
            justify-content: center;
            align-items: center;
            font-weight: 400;
            text-align: center;
        }
    </style>
</head>

<body>
<div class="page">
    {% block content %}
    <h3>{{ exc.status_code }}.</h3>
    <p>{{ exc.detail }}</p>
    {% endblock %}
</div>
</body>
</html>
